@@ -11,9 +11,10 @@ from twisted.python.deprecate import deprecated
 from twisted.python.versions import Version
 
 from foolscap import ipb, base32, negotiate, broker, eventual, storage
-from foolscap import connection, util
-from foolscap.connections.tcp import DefaultTCP
+from foolscap import connection, util, info
+from foolscap.connections import tcp
 from foolscap.referenceable import SturdyRef
+from .furl import BadFURLError
 from foolscap.tokens import PBError, BananaError, WrongTubIdError, \
      WrongNameError, NoLocationError
 from foolscap.reconnector import Reconnector
@@ -32,7 +33,6 @@ class Listener(protocol.ServerFactory, service.Service):
 
     def __init__(self, tub, endpoint_or_description, _test_options={},
                  negotiationClass=negotiate.Negotiation):
-        service.Service.__init__(self)
         assert isinstance(tub, Tub)
         self._tub = tub
 
@@ -95,7 +95,10 @@ class Listener(protocol.ServerFactory, service.Service):
                      addr=(addr.host, addr.port),
                      facility="foolscap.listener")
         proto = self._negotiationClass(logparent=lp)
-        proto.initServer(self)
+        ci = info.ConnectionInfo()
+        ci._set_listener_description(self._describe())
+        ci._set_listener_status("negotiating")
+        proto.initServer(self, ci)
         proto.factory = self
         return proto
 
@@ -105,6 +108,12 @@ class Listener(protocol.ServerFactory, service.Service):
             tub = self._tub
         return (tub, self._redirects.get(tubID))
 
+    def _describe(self):
+        desc = "Listener"
+        if self._lp:
+            desc += " on %s" % str(self._lp.getHost())
+        return desc
+
 def generateSwissnumber(bits):
     bytes = os.urandom(bits/8)
     return base32.encode(bytes)
@@ -152,8 +161,6 @@ class Tub(service.MultiService):
 
     @ivar brokers: maps TubIDs to L{Broker} instances
 
-    @ivar listeners: maps strport to TCPServer service
-
     @ivar referenceToName: maps Referenceable to a name
     @ivar nameToReference: maps name to Referenceable
 
@@ -248,7 +255,7 @@ class Tub(service.MultiService):
         self.brokers = {} # maps TubRef to a Broker that connects to them
         self.reconnectors = []
 
-        self._connectionHandlers = {"tcp": DefaultTCP()}
+        self._connectionHandlers = {"tcp": tcp.default()}
         self._activeConnectors = []
 
         self._pending_getReferences = [] # list of (d, furl) pairs
@@ -485,14 +492,14 @@ class Tub(service.MultiService):
         """Start listening for connections.
 
         @type  what: string
-        @param what: a L{twisted.application.strports} -style description
+        @param what: a L{twisted.internet.endpoints.serverFromString} -style
+                     description
         @param _test_options: a dictionary of options that can influence
                               connection negotiation before the target Tub
                               has been determined
 
         @return: The Listener object that was created. This can be used to
-        stop listening later on, and to figure out which port was allocated
-        when you used a strports specification of 'tcp:0'."""
+        stop listening later on."""
 
         if what in ("0", "tcp:0"):
             warningString = ("Tub.listenOn('tcp:0') was deprecated "
@@ -527,8 +534,23 @@ class Tub(service.MultiService):
     def getShortTubID(self):
         return self.tubID[:4]
 
+    def getConnectionInfoForFURL(self, furl):
+        try:
+            tubref = SturdyRef(furl).getTubRef()
+        except (ValueError, BadFURLError):
+            return None # unparseable FURL
+        return self._getConnectionInfoForTubRef(tubref)
+
+    def _getConnectionInfoForTubRef(self, tubref):
+        if tubref in self.brokers:
+            return self.brokers[tubref].getConnectionInfo()
+        if tubref in self.tubConnectors:
+            return self.tubConnectors[tubref].getConnectionInfo()
+        return None # currently have no established or in-progress connection
+
     def connectorStarted(self, c):
         assert self.running
+        # TODO: why a list? shouldn't there only ever be one TubConnector?
         self._activeConnectors.append(c)
     def connectorFinished(self, c):
         if c in self._activeConnectors:
@@ -884,8 +906,9 @@ class Tub(service.MultiService):
         t1.setPeer(t2); t2.setPeer(t1)
         n = negotiate.Negotiation()
         params = n.loopbackDecision()
-        b1,b2 = (self.brokerClass(tubref, params),
-                 self.brokerClass(tubref, params))
+        ci = info.ConnectionInfo()
+        b1 = self.brokerClass(tubref, params, connectionInfo=ci)
+        b2 = self.brokerClass(tubref, params)
         # we treat b1 as "our" broker, and b2 as "theirs", and we pretend
         # that b2 has just connected to us. We keep track of b1, and b2 keeps
         # track of us.
@@ -893,6 +916,10 @@ class Tub(service.MultiService):
         b2.setTub(self)
         t1.protocol = b1; t2.protocol = b2
         b1.makeConnection(t1); b2.makeConnection(t2)
+        ci._set_connected(True)
+        ci._set_winning_hint("loopback")
+        ci._set_connection_status("loopback", "connected")
+        ci._set_established_at(b1.creation_timestamp)
         self.brokerAttached(tubref, b1, False)
         return b1
 
